from cosmoglobe.sky.components import Component
from cosmoglobe.utils.utils import NsideError, _get_astropy_unit

import astropy.units as u
import healpy as hp
import numpy as np


class Model:
    """The Cosmoglobe Sky Model.

    This class acts as a container for the various components making up the
    sky model and provides methods to simulate the sky emission at a given
    frequency or over a bandpass.

    Parameters
    ----------
    components : list
        A list of `cosmoglobe.sky.Component` objects that are added to the
        model.
    nside : int
        Healpix map resolution. Represents the resolution of the sky model. If
        nside is `None`, nside is set to the nside of the first inserted 
        component. Default: ``None``.

    Attributes
    ----------
    AME : `cosmoglobe.sky.components.AME`
        The AME sky component.
    CMB : `cosmoglobe.sky.components.CMB`
        The CMB sky component.
    dust : `cosmoglobe.sky.components.Dust`
        The dust sky component.
    ff : `cosmoglobe.sky.components.FreeFree`
        The free-free sky component.
    radio : `cosmoglobe.sky.components.Radio`
        The radio sky component.
    synch : `cosmoglobe.sky.components.Synchrotron`
        The synchrotron sky component.
    nside : int
        Healpix map resolution.
    components : dict
        Dictionary of all sky components included in the model.


    Methods
    -------
    __call__
    disable
    enable
    to_nside

    """

    def __init__(self, components=None, nside=None):
        self.nside = nside
        self.components = {}
        self.disabled_components = {}
        if components:
            for component in components:
                self._add_component(component)


    def _add_component(self, component):
        if not issubclass(component.__class__, Component):
            raise TypeError(
                f'{component} is not a subclass of cosmoglobe.sky.Component'
            )

        name = component.label
        if name in self.components:
            raise KeyError(f'component {name} already exists in model')

        if component.diffuse:
            nside = hp.get_nside(component.amp)
            if nside != self.nside:
                if self.nside is None:
                    self.nside = nside
                else:
                    raise NsideError(
                        f'component {name!r} has a reference map at NSIDE='
                        f'{nside}, but model NSIDE is set to {self.nside}'
                    )
        # Explicitly set nside for non diffuse components since most attributes
        # are not stored in maps           
        else:
            try:
                component._set_nside(self.nside)
            except AttributeError:
                pass
            
        setattr(self, name, component)
        self.components[name] = component


    @u.quantity_input(freq=u.Hz, bandpass=(u.Jy/u.sr, u.K, None), 
                      fwhm=(u.rad, u.deg, u.arcmin))
    def __call__(self, freq, bandpass=None, fwhm=0.0*u.rad, output_unit=u.uK):
        r"""Simulates the model emission given a single, or, a set of
        frequencies.

        Optionally, a bandpass profile can be given along with the 
        corresponding frequencies.

        .. math::
            
            \begin{aligned}
            \mathbf{s}_{\mathrm{RJ}}(\nu) &=a_{\mathrm{CMB}} \frac{x^{2} 
            \mathrm{e}^{x}}{\left(\mathrm{e}^{x}-1\right)^{2}} 
            \frac{\left(\mathrm{e}^{x_{0}}-1\right)^{2}}{x_{0}^{2} 
            \mathrm{e}^{x_{0}}}+\\
            &+\boldsymbol{a}_{\mathrm{s}}\left(\frac{\nu}{\nu_{0, 
            \mathrm{~s}}}\right)^{\beta_{\mathrm{s}}}+\\
            &+\boldsymbol{a}_{\mathrm{ff}} \frac{g_{\mathrm{ff}}
            \left(\nu ; T_{e}\right)}{g_{\mathrm{ff}}\left(\nu_{0, 
            \mathrm{ff}} ; T_{e}\right)}\left(\frac{\nu_{0, 
            \mathrm{ff}}}{\nu}\right)^{2}+\\
            &+\boldsymbol{a}_{\mathrm{sd}}\left(\frac{\nu_{0, 
            \mathrm{sd}}}{\nu}\right)^{2} \frac{s_{0}^{\mathrm{sd}}
            \left(\nu \cdot \frac{\nu_{p}}{30.0 \mathrm{GHz}}\right)}
            {s_{0}^{\mathrm{sd}}\left(\nu_{0, \mathrm{sd}} \cdot 
            \frac{\nu_{p}}{30.0 \mathrm{GHz}}\right)}+\\
            &+\boldsymbol{a}_{\mathrm{d}}\left(\frac{\nu}{\nu_{0, 
            \mathrm{~d}}}\right)^{\beta_{\mathrm{d}}+1} 
            \frac{\mathrm{e}^{h \nu_{0, \mathrm{~d}} 
            / k T_{\mathrm{d}}}-1}{\mathrm{e}^{\mathrm{h} \nu 
            / k T_{\mathrm{d}}}-1}+\\&+\sum_{j=1}^{N_{\mathrm{scc}}} 
            \boldsymbol{a}_{\mathrm{src}}^{j}\left(\frac{\nu}{\nu_{0, 
            \mathrm{src}}}\right)^{\alpha_{j, \mathrm{src}}-2}
            \end{aligned}



        Parameters
        ----------
        freq : `astropy.units.Quantity`
            A frequency, or list of frequencies for which to evaluate the
            sky emission.
        bandpass : `astropy.units.Quantity`
            Bandpass profile corresponding to the frequencies. If None, a
            delta peak in frequency is assumed at the given frequencies.
            Default: None
        output_unit : `astropy.units.Unit`
            The desired output units of the emission. Must be signal units. 
            Default : uK

        Returns
        -------
        astropy.units.Quantity
            Model emission.
        """

        if bandpass is None and freq.ndim > 0:
            return [
                self._get_model_emission(freq, bandpass, fwhm, output_unit)
                for freq in freq
            ]

        return self._get_model_emission(freq, bandpass, fwhm, output_unit)


    def _get_model_emission(self, freq, bandpass, fwhm, output_unit):
        if self.is_polarized:
            shape = (3, hp.nside2npix(self.nside))
        else:
            shape = (1, hp.nside2npix(self.nside))
        diffuse_emission = np.zeros(shape)
        ptsrc_emission = np.zeros(shape)

        unit = _get_astropy_unit(output_unit)
        diffuse_emission = u.Quantity(diffuse_emission, unit=unit)
        ptsrc_emission = u.Quantity(ptsrc_emission, unit=unit)

        for comp in self:
            if comp.diffuse:
                comp_emission = comp(freq, bandpass, output_unit=output_unit)
                for idx, col in enumerate(comp_emission):
                    diffuse_emission[idx] += col
            else:
                comp_emission = comp(freq, bandpass, fwhm=fwhm, output_unit=output_unit)
                for idx, col in enumerate(comp_emission):
                    ptsrc_emission[idx] += col

        if fwhm is not None:
            # If diffuse emission is non-zero
            print('Smoothing diffuse emission')
            if diffuse_emission.value.any():
                diffuse_emission = hp.smoothing(
                    diffuse_emission, fwhm.to(u.rad).value
                ) * diffuse_emission.unit

        return diffuse_emission + ptsrc_emission


    def disable(self, component):
        """Disable a component in the model.

        Parameters
        ----------
        component : str, `cosmoglobe.sky.Component`
            The name of a component or the the component class in the model.
        """
        
        if isinstance(component, str):
            comp = component
        elif isinstance(component.__class__, Component):
            comp = component.label
        else:
            raise ValueError(
                'component must be the component label in the model or the '
                'component object'
            )
        try:
            self.disabled_components[comp] = self.components[comp]
        except KeyError:
            raise KeyError(f'{comp} is not enabled')
        del self.components[comp]


    def enable(self, component):
        """enable a disabled component.

        Parameters
        ----------
        component : str, `cosmoglobe.sky.Component`
            The name of a component or the the component class in the model.
        """

        if isinstance(component, str):
            comp = component
        elif isinstance(component.__class__, Component):
            comp = component.label
        else:
            raise ValueError(
                'component must be the component label in the model or the '
                'component object'
            )

        try:
            self.components[comp] = self.disabled_components[comp]
        except KeyError:
            raise KeyError(f'{comp} is not disabled')
        del self.disabled_components[comp]


    def _insert_component(self, component):
        """Insert a new component to the model.

        Parameters
        ----------
        component : `cosmoglobe.sky.Component`:
            Sky component to be added to the model. Must be a subclass of 
            `cosmoglobe.sky.Component`.

        """
        self._add_component(component)


    def _remove_component(self, name):
        """Removes a component from the model.

        Parameters
        ----------
        name : str
            Component attribute name.
        """

        del self[name]


    def to_nside(self, new_nside):
        """ud_grades all maps in the component to a new nside.

        Parameters
        ----------
        new_nside : int
            Healpix map resolution parameter.
        """

        if new_nside == self.nside:
            print(f'Model is already at nside {new_nside}')
            return
        if not hp.isnsideok(new_nside, nest=True):
            raise ValueError(f'nside: {new_nside} is not valid.')
        
        self.nside = new_nside
        for comp in self:
            comp.to_nside(new_nside)




    @property
    def is_polarized(self):
        """Returns True if model includes a polarized component and False 
        otherwise.
        """
        for comp in self:
            if comp.is_polarized:
                return True
        return False


    def __iter__(self):
        return iter(self.components.values())


    def __len__(self):
        return len(self.components)


    def __delitem__(self, name):
        if name not in self.components:
            raise KeyError(f'component {name} does not exist')
        delattr(self, name)
        del self.components[name]


    def __repr__(self):
        reprs = []
        for key, component in self.components.items():
            component_repr = repr(component) + '\n'
            reprs.append(f'({key}): {component_repr}')

        main_repr = f'Model('
        main_repr += f'\n  nside: {self.nside}'
        main_repr += '\n  components( '
        main_repr += '\n    ' + '    '.join(reprs)
        main_repr += f'  )'
        main_repr += f'\n)'

        return main_repr


from .constants import h, c, k_B, T_0
import numpy as np
import astropy.units as u

def blackbody_emission(freq, T):
    """Returns the emission emitted by a blackbody with with temperature T at 
    a frequency freq in SI units: W/(m^2 Hz sr).

    Args:
    -----
    freq (astropy.units.quantity.Quantity):
        Frequency in units of Hertz.
    T (astropy.units.quantity.Quantity):
        Temperature of the blackbody in units of K. 

    Returns:
    --------
    (astropy.units.quantity.Quantity):
        Blackbody emission in units in SI units.

    """
    try:
        T = T.astype(np.float64)
    except AttributeError:
        pass

    emission = ((2*h*freq**3)/c**2) / np.expm1((h*freq)/(k_B*T)) / u.sr
    return emission.to(u.W/(u.m**2 * u.Hz * u.sr))


def gaunt_factor(freq, T_e):
    """Returns the gaunt factor at a given frequency and electron temperature
    in SI units.

    Args:
    -----
    freq (astropy.units.quantity.Quantity):
        Frequency in units of Hertz.   
    T_e (astropy.units.quantity.Quantity):
        Electron temperature in units of K.

    Returns:
    --------
    (astropy.units.quantity.Quantity):
        Gaunt Factor.

    """
    try:
        T_e = T_e.astype(np.float64)
    except AttributeError:
        pass

    try:
        freq = (freq.to(u.GHz)).value   
        T_e = (T_e.to(u.kK).value)/10
    except AttributeError:
        pass    
        
    gaunt_factor = np.log(np.exp(5.96 - (np.sqrt(3)/np.pi) * np.log(freq
                  * (T_e)**-1.5)) + np.e)

    return u.Quantity(gaunt_factor)



def thermodynamical_to_brightness(freq, T=T_0):
    """Returns the conversion factor between thermodynamical and brightness 
    temperatures (K_CMB and K_RJ).

    Args:
    -----
    freq (astropy.units.quantity.Quantity):
        Frequency in units of Hertz.   

    Returns:
    --------
    (astropy.units.quantity.Quantity):
        K_CMB -> K_RJ factor.

    """  
    x = (h*freq) / (k_B*T)
    return ((x**2 * np.exp(x)) / (np.expm1(x)**2)).si


def brightness_to_thermodynamical(freq):
    """Returns the conversion factor between brightness and thermodynamical 
    temperatures (K_CMB and K_RJ).

    Args:
    -----
    freq (astropy.units.quantity.Quantity):
        Frequency in units of Hertz.   

    Returns:
    --------
    (astropy.units.quantity.Quantity):
        K_RJ -> K_CMB factor.

    """  
    return 1/thermodynamical_to_brightness(freq)


from cosmoglobe.utils.constants import h, c, k_B, T_0

from scipy.interpolate import RectBivariateSpline
import astropy.units as u
import numpy as np

@u.quantity_input(bandpass=(u.Jy/u.sr, u.K), freqs=u.Hz)
def get_normalized_bandpass(bandpass, freqs, unit=u.K):
    """Returns a bandpass profile normalized to unity when integrated over all 
    frequencies corresponding to the profile in units of K_RJ.

    Parameters
    ----------
    bandpass: `astropy.units.Quantity`
        Bandpass profile in signal units.    
    freqs: `astropy.units.Quantity`
        Frequencies corresponding to the bandpass profile.
    unit: `astropy.units.UnitBase`
        The unit we want to convert to.
    """
    bandpass = bandpass.to(
        unit=unit, equivalencies=u.brightness_temperature(freqs)
    )

    return bandpass/np.trapz(bandpass, freqs)


@u.quantity_input(bandpass=u.Hz**-1, freqs=u.Hz)
def get_bandpass_coefficient(bandpass, freqs, output_unit):
    """Computes the bandpass coefficient for a map that has been calibrated in 
    K_RJ to some output unit, after taking into account the bandpass.

    Parameters
    ----------
    bandpass: `astropy.units.Quantity`
        Normalized bandpass profile in units of 1/Hz.    
    freqs: `astropy.units.Quantity`
        Frequencies corresponding to the bandpass profile.
    unit: `astropy.units.UnitBase`
        The unit we want to convert to.

    Returns
    -------
    bandpass_coefficient: `astropy.units.Quantity`
        The bandpass coefficient.
        
    """
    intensity_derivative_i = b_rj(freqs)

    #Selecting the intensity derivative expression given an output_unit
    try:
        output_unit = u.Unit(output_unit)
        if u.K in output_unit.si.bases:
            intensity_derivative_j = b_rj(freqs)
        elif output_unit == u.MJy/u.sr:
            intensity_derivative_j = b_iras(freqs, np.mean(freqs))

    except ValueError:
        if isinstance(output_unit, str):
            if output_unit.lower().endswith('k_rj'):
                intensity_derivative_j = b_rj(freqs)
            elif output_unit.lower().endswith('k_cmb'):
                intensity_derivative_j = b_cmb(freqs)
            else:
                raise NotImplementedError(
                    f'output_unit {output_unit} is not implemented'
                )
    bandpass_coefficient = (
        np.trapz(bandpass*intensity_derivative_i, freqs)/
        np.trapz(bandpass*intensity_derivative_j, freqs)
    )

    return bandpass_coefficient


def get_interp_parameters(spectral_parameters):
    """Returns the interpolation range of the spectral parameters of a 
    sky component. We use a regular grid with n points for the range.
    
    Parameters
    ----------
    spectral_parameters: dict
        Dictionary containing the spectral parameters of a given component.

    Returns
    -------
    interp_parameters: dict
        Dictionary with a interpolation grid for each spatially varying
        spectral parameter

    """
    dim = 0
    for value in spectral_parameters.values():
        if value.size > 3:
            dim += 1
        
    interp_parameters = {}
    if dim == 0:
        return interp_parameters
    elif dim == 1:
        n = 1000
    elif dim == 2:
        n = 100

    for key, value in spectral_parameters.items():
        if value.size > 3:
            min_, max_ = np.amin(value), np.amax(value)
            interp = np.linspace(min_, max_, n)
            interp_parameters[key] = interp

    return interp_parameters


def interp1d(comp, freqs, bandpass, interp_parameter, spectral_parameters):
    """Performs a 1-dimensional bandpass integration (numpy interp1d) for a 
    sky component and returns the frequency scaling factor.
    
    Parameters:
    -----------
    comp: `cosmoglobe.sky.Component`
        A Cosmoglobe sky component object.
    freqs: `astropy.units.Quantity`
        Frequencies corresponding to the bandpass profile.
    bandpass: `astropy.units.Quantity`
        Normalized bandpass profile in units of 1/Hz.    
    interp_parameter: `astropy.units.Quantity`
        Spectral parameter *grid* to interpolate over.
    spectral_parameters: dict
        Dictionary containing *all* the spectral parameters of the component. 

    Returns:
    --------
    scaling: `astropy.units.Quantity`
        Frequency scaling factor obtained by integrating over the bandpass.
        
    """
    integrals = []
    for key, interp_param in interp_parameter.items():
        for grid_point in interp_param:
            spectral_parameters[key] = grid_point
            freq_scaling = comp.get_freq_scaling(freqs, comp.freq_ref, 
                                                 **spectral_parameters)

            integrals.append(np.trapz(freq_scaling*bandpass, freqs))

        #Reshape integrals list to (3, n) 
        integrals = u.Quantity(np.transpose(integrals), unit=integrals[0].unit)
        if integrals.ndim == 1:
            integrals = np.expand_dims(integrals, axis=0)

        # Find integrated scaling factor for IQU
        if len(comp.spectral_parameters[key]) == 3:
            scaling =  [
                np.interp(comp.spectral_parameters[key][idx].value, 
                interp_param.value, col.value)
                for idx, col in enumerate(integrals)
            ]

        # Find integrated scaling factor for I
        else:
            scaling =  [
                np.interp(comp.spectral_parameters[key].value, 
                interp_param.value, col.value) for col in integrals
            ]
        return scaling


def interp2d(comp, freqs, bandpass, interp_parameters, spectral_parameters):
    """Performs a 2-dimensional bandpass integration 
    (scipy RectBivariateSpline) for a sky component and returns the frequency 
    scaling factor.
    
    Parameters
    ----------
    comp: `cosmoglobe.sky.Component`
        A Cosmoglobe sky component object.
    freqs: `astropy.units.Quantity`
        Frequencies corresponding to the bandpass profile.
    bandpass: `astropy.units.Quantity`
        Normalized bandpass profile in units of 1/Hz.    
    interp_parameters: `astropy.units.Quantity`
        Spectral parameters *grids* to interpolate over.
    spectral_parameters: dict
        Dictionary containing *all* the spectral parameters of the component. 

    Returns
    -------
    scaling: `astropy.units.Quantity`
        Frequency scaling factor obtained by integrating over the bandpass.
        
    """
    # Make n x n mesh grid for the spectral parameters
    n = len(list(interp_parameters.values())[0])
    grid = {
        key: value for key, value in zip(
            interp_parameters.keys(), np.meshgrid(*interp_parameters.values())
        )
    }

    integrals = np.zeros((n, n, 3)) if comp.is_polarized else np.zeros((n, n, 1))
    for i in range(n):
        for j in range(n):
            grid_spectrals = {key: value[i,j] for key, value in grid.items()}
            freq_scaling = comp.get_freq_scaling(
                freqs, comp.freq_ref, **grid_spectrals
            )
            integrals[i,j] = np.trapz(freq_scaling*bandpass, freqs)
    integrals = np.transpose(integrals)

    scaling = []
    for idx, col in enumerate(integrals):
        f = RectBivariateSpline(*interp_parameters.values(), col)
        packed_spectrals = [
            spectral[idx] if spectral.shape[0] == 3 else spectral[0]
            for spectral in spectral_parameters.values()
        ]
        scaling.append(f(*packed_spectrals, grid=False))
    
    return scaling



# Intensity derivatives
def b_rj(freq):
    """The intensity derivative in unit convention K_RJ.

    Args:
    -----
    freq (astropy.units.quantity.Quantity):
        Frequency in units of Hertz.   

    Returns:
    --------
    (astropy.units.quantity.Quantity):
        Jy/sr -> K_RJ factor.

    """
    return (2*k_B*freq**2)/(c**2 * u.sr)


def b_cmb(freq, T=T_0):
    """The intensity derivative in unit convention K_CMB.

    Parameters:
    -----------
    freq (astropy.units.quantity.Quantity):
        Frequency in units of Hertz.   
    T (astropy.units.quantity.Quantity):
        The CMB emperature. Default is T_0.

    Returns:
    --------
    (astropy.units.quantity.Quantity):
        Bandpass coefficient

    """
    x = (h*freq)/(k_B*T)
    return (
        (2*h*freq**3)/(c**2 * np.expm1(x)) * 
        (np.exp(x)/np.expm1(x))*((h*freq)/(k_B*T**2)) * u.sr**-1
    )


def b_iras(freq, freq_ref):
    """The intensity derivative in the IRAS unit convention (MJy/sr).

    Parameters:
    -----------
    freq (astropy.units.quantity.Quantity):
        Frequency in units of Hertz.   

    Returns:
    --------
    (astropy.units.quantity.Quantity):
        Bandpass coefficient

    """
    return (freq_ref/freq)


import numpy as np
import astropy.units as u

class NsideError(Exception):
    """Raised if there is a NSIDE related problem"""
    pass

class ModelError(Exception):
    """Raised if there is a Model related problem"""
    pass


def _extract_scalars(spectral_parameters):
    """Extracts scalars from a spectral_parameter dict. If a map is constant
    in all dims, then it is converted to a (ndim,) array"""
    scalars = {}
    for key, value in spectral_parameters.items():
        if value.size > 1:
            uniques = [np.unique(col) for col in value.value]
            all_is_unique = all([len(col) == 1 for col in uniques])
            if all_is_unique:
                scalars[key] = uniques*value.unit

        else:
            scalars[key] = value
    return scalars


def _set_spectral_units(maps):
    #TODO: Figure out how to correctly detect unit of spectral map in chain.
    #      Until then, a hardcoded dict is used:
    units = dict(
        T=u.K,
        Te=u.K,
        nu_p=u.GHz,
    )
    for map_ in maps:
        if map_ in units:
            maps[map_] *= units[map_]
        else:
            maps[map_] *= u.dimensionless_unscaled

    return maps


def _get_astropy_unit(unit):
    try:
        output_unit = u.Unit(unit)
    except ValueError:
        if unit.lower().endswith('k_rj'):
            output_unit = u.Unit(unit[:-3])
            
        elif unit.lower().endswith('k_cmb'):
            output_unit = u.Unit(unit[:-4])

    return output_unit


def gaussian_beam_2D(r, sigma):
    """Returns the Gaussian beam in 2D in polar coordinates.

    Parameters:
    -----------
    r : float
        Angular distance.
    sigma : float
        The sigma of the Gaussian (beam radius).
    
    """
    return r*np.exp(-(r**2)/(2 * sigma**2))/(sigma*np.sqrt(2*np.pi))



from cosmoglobe.sky.model import Model
from cosmoglobe.sky import components
from cosmoglobe.utils.utils import ModelError

from astropy.utils.data import download_file
from astropy.io.misc import fnpickle, fnunpickle
import numpy as np

data_url = 'http://cosmoglobe.uio.no/BeyondPlanck/precomputed/'

COSMOGLOBE_COMPS = dict(
    dust=components.Dust,
    synch=components.Synchrotron,
    ff=components.FreeFree,
    ame=components.AME,
    cmb=components.CMB,
    radio=components.Radio
)


def save_model(model, filename):
    """Saves a model to file in form of a dictionary.
    
    Args:
    -----
    model (cosmoglobe.sky.Model):
        The sky model to save.
    filename (str):
        Save filename.

    """

    model_dict = {}
    try:
        for comp in model:
            freq_ref =  comp.freq_ref
            if freq_ref is not None:
                if freq_ref.ndim > 0:
                    # freq_ref is stored as (3,1) or (1,1) arrays in model for broadcasting
                    freq_ref = np.squeeze(freq_ref)
                    freq_ref = [freq_ref[0].value, freq_ref[-1].value]*freq_ref.unit
                else:
                    freq_ref = freq_ref
            
            comp_dict = {
                comp.name : {
                    'amp': comp.amp,
                    'freq_ref': freq_ref,
                    'spectral_parameters': comp.spectral_parameters,
                }
            }
            model_dict.update(comp_dict)

    except AttributeError:
        raise ModelError(
            'Model is not compatible with the current Cosmoglobe Sky Model'
        )

    fnpickle(model_dict, filename)


def load_model(path_to_model):
    """Loads a model from file.
    
    Args:
    -----
    path_to_model (str):
        Path to the model file.

    Returns:
    --------
    model (cosmoglobe.sky.Model):
        Loaded cosmoglobe sky model.

    """
    model_dict =  fnunpickle(path_to_model)

    model = model_from_dict(model_dict)

    return model



def model_from_dict(model_dict):
    """Creates and returns a cosmoglobe.sky.Model from a model dictionary."""
    model = Model()
    for key, value in model_dict.items():
        try:
            comp = COSMOGLOBE_COMPS[key]
        except KeyError:
            raise ModelError(
                'Model components does not match the Cosmoglobe Sky Model'
            )

        try:
            amp = value['amp']
            freq_ref = value['freq_ref']
            spectral_parameters = value['spectral_parameters']
        except KeyError as e:
            raise e

        if freq_ref is not None: #CMB is initialized without a reference freq
            model.insert(comp(key, amp, freq_ref, **spectral_parameters))    
        else:
            model.insert(comp(key, amp, **spectral_parameters))    
        
    return model

    

def _download_BP_model(release: int, nside: int, cache: bool = True) -> Model:
    model_name = f'BP_test_model_r{release}_n{nside}.pkl'
    path_to_model = download_file(data_url + model_name, cache=cache)

    with open(path_to_model, 'rb') as f:
        return pickle.load(f)


def BP(release: int , nside: int, cache: bool = True) -> Model:
    """Loads the BeyondPlanck sky model for a given BP release. 

    The model is downloaded and cached using astropy.utils.

    Args:
    -----
    release (int):
        BeyondPlanck release number.
    nside (int):
        Healpix resolution parameter. Model is downloaded at the given nside.
    cache (bool):
        If True, the downloaded model is cached away for later used.

    Returns:
    (cosmoglobe.sky.Model):
        The BeyondPlanck sky model at a given nside for a release.
        
    """
    model = _download_BP_model(release, nside, cache=cache)
    return model



from cosmoglobe.utils.utils import ModelError
from logging import error
from cosmoglobe.sky.model import Model

import warnings
import os
import healpy as hp
import numpy as np
import astropy.units as u
import matplotlib.pyplot as plt
from .plottools import *

# Fix for macos openMP duplicate bug
os.environ["KMP_DUPLICATE_LIB_OK"] = "True"

def plot(
    input,
    sig=0,
    comp=None,
    freq=None,
    ticks=None,
    min=None,
    max=None,
    cbar=True,
    unit=None,
    coord=None,
    graticule=False,
    projection_type="mollweide",
    fwhm=0.0,
    nside=None,
    mask=None,
    cmap=None,
    norm=None,
    remove_dip=False,
    remove_mono=False,
    title=None,
    ltitle=None,
    width="m",
    darkmode=False,
    interactive=False,
    **kwargs,
):
    """
    General plotting function for maps.
    This function is a wrapper on healpys projview function with some added features.
    Added features include lognorm, 

    TODO:
    Subplots and norm cleanup

    Parameters
    ----------
    input : ndarray, fits file path or cosmoglobe model object
        Map data input given as numpy array either 1d or index given by "sig".
        Also supports fits-file path string or cosmoglobe model.
        If cosmoglobe object is passed such as "model", specify comp or freq.
    sig : list of int or string ["I", "Q", "U"]
        Signal indices to be plotted. 0, 1, 2 interprated as IQU.
        default = [0,]
    comp : string
        Component label for automatic identification of plotting
        parameters based on information from autoparams.json default = None
    freq : float
        frequency in GHz needed for scaling maps when using a model object input
    ticks : list or str
        Min and max value for data. If None, uses 97.5th percentile.
        default = None
    cbar : bool
        Adds a cbar, and "cb" to output filename.
        cbar = False
    graticule : cbar
        Adds graticule to figure.
        default = False
    fwhm : float
        Optional map smoothing. FWHM of gaussian smoothing in arcmin.
        default = 0.0
    mask : array of healpix mask
        Apply a mask file to data
        default = None
    remove_dip : bool
        If mdmask is specified, fits and removes a dipole.
        default = True
    remove_mono : bool
        If mdmask is specified, fits and removes a monopole.
        default = True
    norm : str
        if norm=="linear":
            normal 
        if norm=="log":
            Normalizes data using a semi-logscale linear between -1 and 1.
            Autodetector uses this sometimes, you will be warned.
        default = None
    darkmode : bool
        Plots all outlines in white for dark backgrounds, and adds "dark" in
        filename.
        default = False
    cmap : str
        Colormap (ex. sunburst, planck, jet). Both matplotliib and cmasher
        available as of now. Also supports qualitative plotly map, [ex.
        q-Plotly-4 (q for qualitative 4 for max color)] Sets planck as default.
        default = None
    title : str
        Sets the upper right title. Has LaTeX functionaliity (ex. $A_{s}$.)
        default = None
    ltitle : str
        Sets the upper left title. Has LaTeX functionaliity (ex. $A_{s}$.)
        default = None
    """
    # Pick sizes from size dictionary for page width plots
    if isinstance(width, str):
        width = {"x": 2.75, "s": 3.5, "m": 4.7, "l": 7,}[width]
    height = width / 2
    if cbar:
        height *= 1.275  # Size correction with cbar
    figratio = height / width
    set_style(darkmode)

    # Currently not working with projview
    """
    # Make figure
    fig, ax = make_fig(
        (width, height),
        fignum,
        hold,
        subplot,
        reuse_axes, 
        projection="mollweide",
    )
    """

    # Translate sig to correct format
    stokes = [
        "I",
        "Q",
        "U",
    ]
    if isinstance(sig, str):
        sig = stokes.index(sig)

    # Parsing component string
    comp_full = comp
    if comp is not None:
        comp, *specparam = comp.split()
    diffuse = True # For future functionality

    # If map is string, interprate as file path
    if isinstance(input, str):
        m = hp.read_map(input, field=sig)
    elif isinstance(input, Model):
        """
        Get data from model object with frequency scaling        
        """
        if comp == None:
            if freq is not None:
                m = input(freq * u.GHz, fwhm=fwhm * u.arcmin,)
            else:
                raise ModelError(
                    f"Model object passed with comp and freq set to None"
                    f"comp: {comp} freq: {freq}"
                )
        else:
            if len(specparam) > 0 and isinstance(specparam[0], str):
                m = getattr(input, comp).spectral_parameters[specparam[0]]

                if len(m[sig]) == 1:
                    warnings.warn(
                        "Same value across the whole sky, mapping to array of length Npix"
                    )
                    m = np.full(hp.nside2npix(input.nside), m[sig])
            else:
                if freq == None:
                    freq = getattr(input, comp).freq_ref.value
                    m = getattr(input, comp).amp
                    if comp == "radio":
                        m = getattr(input, comp).get_map(m, fwhm=fwhm*u.arcmin)
                        diffuse = False
                    try:
                       freq = round(freq.squeeze()[sig], 5)
                    except IndexError:
                       freq = round(freq, 5)
                else:
                    m = getattr(input, comp)(freq * u.GHz, fwhm=fwhm * u.arcmin,)
        if isinstance(m, u.Quantity):
            m = m.value
    else:
        if isinstance(input, np.ndarray):
            m = input
        else:
            raise TypeError(
                f"Type {type(input)} of input not supported"
                f"Supports numpy array, cosmoglobe model object or fits file string"
            )

    # Make sure we have 1d array at this point
    m = m[sig] if m.ndim > 1 else m
    # Mask map
    if mask is not None:
        m = hp.ma(m)
        m.mask = np.logical_not(mask)

    # udgrade map
    if nside != None and nside != hp.get_nside(m):
        m = hp.ud_grade(m, nside)
    else:
        nside = hp.get_nside(m)

    # Smooth map
    if fwhm > 0.0 and diffuse:
        m = hp.smoothing(m, (fwhm * u.arcmin).to(u.rad).value)

    # Remove mono/dipole
    if remove_dip:
        m = hp.remove_dipole(m, gal_cut=30, copy=True, verbose=True)
    if remove_mono:
        m = hp.remove_monopole(m, gal_cut=30, copy=True, verbose=True)

    # Fetching autoset parameters
    params = autoparams(
        comp_full, sig, title, ltitle, unit, ticks, min, max, norm, cmap, freq
    )
    # Ticks and ticklabels
    ticks = params["ticks"]
    if ticks == "auto":
        ticks = get_percentile(m, 97.5)
    elif None in ticks:
        pmin, pmax = get_percentile(m, 97.5)
        if ticks[0] == None:
            ticks[0] = pmin
        if ticks[-1] == None:
            ticks[-1] = pmax

    # Special case if dipole is detected in freqmap
    if freq != None and comp == None:
        # if colorbar is super-saturated
        if len(m[abs(m)>ticks[-1]])/hp.nside2npix(nside) > 0.7:
            ticks = [tick*10 for tick in ticks]
    ticklabels = [fmt(i, 1) for i in ticks]

    # Math text in labels
    for i in ["title", "unit", "left_title"]:
        if params[i] and params[i] != "":
            params[i] = r"$" + params[i] + "$"

    #### Color map #####
    cmap = load_cmap(params["cmap"], params["norm"])
    #### Logscale ####
    if params["norm"] == "log":
        m, ticks = apply_logscale(m, ticks, linthresh=1)

    # Plot using mollview if interactive mode
    if interactive:
        if params["norm"]=="log":
            params["norm"] = None
            # Ticklabels not available for this
            params["unit"] = r'$\log($' +params["unit"]+r"$)$"
        ret = hp.mollview(
            m,
            min=ticks[0],
            max=ticks[-1],
            cbar=cbar,
            cmap=cmap,
            coord=coord,
            unit=params["unit"],
            title=params["title"],
            norm=params["norm"],
            **kwargs,
        )
        return


    # Plot figure
    ret = hp.newvisufunc.projview(
        m,
        min=ticks[0],
        max=ticks[-1],
        cbar=False,
        cmap=cmap,
        projection_type=projection_type,
        graticule=graticule,
        override_plot_properties={
            "figure_width": width,
            "figure_size_ratio": figratio,
        },
        coord=coord,
        fontsize={
            "xlabel": 8,
            "ylabel": 8,
            "xtick_label": 8,
            "ytick_label": 8,
            "title": 8,
            "cbar_label": 8,
            "cbar_tick_label": 8,
        },
        **kwargs,
    )
    plt.gca().collections[-1].colorbar.remove()

    # Remove color bar because of healpy bug
    # Add pretty color bar
    if cbar:
        apply_colorbar(
            plt.gcf(),
            plt.gca(),
            ret,
            ticks,
            ticklabels,
            params["unit"],
            linthresh=1,
            norm=params["norm"],
        )

    #### Right Title ####
    plt.text(
        4.5, 1.1, params["title"], ha="center", va="center",
    )
    #### Left Title (stokes parameter label by default) ####
    plt.text(
        -4.5, 1.1, params["left_title"], ha="center", va="center",
    )


def apply_colorbar(fig, ax, image, ticks, ticklabels, unit, linthresh, norm=None):
    """
    This function applies a colorbar to the figure and formats the ticks.
    """
    from matplotlib.ticker import FuncFormatter

    cb = fig.colorbar(
        image,
        ax=ax,
        orientation="horizontal",
        shrink=0.4,
        pad=0.04,
        ticks=ticks,
        format=FuncFormatter(fmt),
    )
    cb.ax.set_xticklabels(ticklabels)
    cb.ax.xaxis.set_label_text(unit)
    if norm == "log":
        linticks = np.linspace(-1, 1, 3) * linthresh
        logmin = np.round(ticks[0])
        logmax = np.round(ticks[-1])

        logticks_min = -(10 ** np.arange(0, abs(logmin) + 1))
        logticks_max = 10 ** np.arange(0, logmax + 1)
        ticks_ = np.unique(np.concatenate((logticks_min, linticks, logticks_max)))
        # cb.set_ticks(np.concatenate((ticks,symlog(ticks_))), []) # Set major ticks

        logticks = symlog(ticks_, linthresh)
        logticks = [x for x in logticks if x not in ticks]
        cb.set_ticks(np.concatenate((ticks, logticks)))  # Set major ticks
        cb.ax.set_xticklabels(ticklabels + [""] * len(logticks))

        minorticks = np.linspace(-linthresh, linthresh, 5)
        minorticks2 = np.arange(2, 10) * linthresh

        for i in range(len(logticks_min)):
            minorticks = np.concatenate((-(10 ** i) * minorticks2, minorticks))
        for i in range(len(logticks_max)):
            minorticks = np.concatenate((minorticks, 10 ** i * minorticks2))

        minorticks = symlog(minorticks, linthresh)
        minorticks = minorticks[(minorticks >= ticks[0]) & (minorticks <= ticks[-1])]
        cb.ax.xaxis.set_ticks(minorticks, minor=True)

    cb.ax.tick_params(
        which="both", axis="x", direction="in",
    )
    cb.ax.xaxis.labelpad = 0
    # workaround for issue with viewers, see colorbar docstring
    cb.solids.set_edgecolor("face")
    return cb



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="cosmoglobe",
    version="0.8.99",
    author="Metin San",
    author_email="metinisan@gmail.com",
    description="The Cosmoglobe Sky Model",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Cosmoglobe/Cosmoglobe",
    packages=setuptools.find_packages(),
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
    install_requires=[
        "healpy",
        "astropy",
        "numpy",
        "h5py>=3.0",
        "scipy",
        "numba",
        "cmasher",
        "tqdm",
    ]
)


from cosmoglobe.utils.bandpass import (
    get_bandpass_coefficient,
    get_interp_parameters, 
    get_normalized_bandpass, 
    interp1d,
    interp2d,
)
from cosmoglobe.utils.functions import (
    blackbody_emission,
    brightness_to_thermodynamical, 
    gaunt_factor, 
    thermodynamical_to_brightness
)
from cosmoglobe.utils.utils import _get_astropy_unit, gaussian_beam_2D

from pathlib import Path
from sys import exit
import warnings
from tqdm import tqdm
import astropy.units as u
import numpy as np
import healpy as hp
import sys

DATA_DIR = Path(__file__).resolve().parent.parent / 'data'


class Component:
    """Base class for a sky component from the Cosmoglobe Sky Model.

    A component is defined by its get_freq_scaling function, which it is 
    required to implement. This function is required to take in a freq and a 
    freq_ref at minimum (even if it is not required to evalute the emission). 
    Further, the get_freq_scaling function can take in any number of spectral 
    parameters, although a model with more than two spectral parameters that 
    varies across the sky is currently not supported under bandpass integration.

    Args:
    -----
    amp (`astropy.units.Quantity`):
        Emission templates of the component at the reference frequencies given
        by freq_ref.
    freq_ref (`astropy.units.Quantity`):
        Reference frequencies for the amplitude template in units of Hertz.
        The input must be an astropy quantity containing the the reference 
        frequency for the stokes I amplitude template, and optionally the
        reference frequency for the soktes Q and U templates if the component 
        is polarized. Example: freq_ref=freq_ref_I*u.GHz, or 
        freq_ref=[freq_ref_I, freq_ref_P]*u.GHz
    spectral_parameters (dict):
        Spectral parameters required to compute the frequency scaling factor. 
        These can be scalars, numpy ndarrays or astropy quantities. 
        Default: None

    """
    def __init__(self, amp, freq_ref, **spectral_parameters):
        self.amp = amp
        self.freq_ref = self._set_freq_ref(freq_ref)
        self.spectral_parameters = spectral_parameters


    @staticmethod
    @u.quantity_input(freq=u.Hz)
    def _set_freq_ref(freq_ref):
        """Reshapes the freq_ref into a broadcastable shape"""
        if freq_ref is None:
            return
        elif freq_ref.ndim == 0:
            return freq_ref
        elif freq_ref.ndim == 1:
            return np.expand_dims(
                u.Quantity([freq_ref[0], freq_ref[1], freq_ref[1]]), axis=1
            )
        else:
            raise ValueError('freq_ref must have a maximum len of 2')


    @u.quantity_input(freq=u.Hz, bandpass=(u.Jy/u.sr, u.K, None), 
                      fwhm=(u.rad, u.deg, u.arcmin))
    def __call__(self, freq, bandpass=None, fwhm=0.0*u.rad, output_unit=u.uK):
        r"""Computes the simulated component emission at an arbitrary frequency 
        or integrated over a bandpass.

        .. math::

            \mathbf{s}_\mathrm{comp} = \mathbf{a}_\mathrm{comp} \; 
            \mathrm{scaling}_\mathrm{comp}

        where :math:`\mathbf{a}_\mathrm{comp}` is the amplitude template of 
        the component at some reference frequency, and 
        :math:`\mathrm{scaling}_\mathrm{comp}` is the scaling factor for 
        component.


        Parameters
        ----------
        freq : `astropy.units.Quantity`
            A frequency, or a list of frequencies at which to evaluate the 
            component emission. If a corresponding bandpass is not supplied, 
            a delta peak in frequency is assumed.
        bandpass : `astropy.units.Quantity`
            Bandpass profile in signal units. The shape of the bandpass must
            match that of the freq input. Default : None
        output_unit : `astropy.units.Unit`
            The desired output unit of the emission. Must be signal units. 
            Default: None
        fwhm : `astropy.units.Quantity`
            The full width half max parameter of the beam. Default: None

        Returns
        -------
        `astropy.units.Quantity`
            Component emission.

        """
        freq_ref = self.freq_ref
        input_unit = u.Unit('uK')

        # Expand dimension on rank-1 arrays from from (n,) to (n, 1) to support
        # broadcasting with (1, nside) or (3, nside) arrays
        amp = self.amp if self.amp.ndim != 1 else np.expand_dims(self.amp, axis=0)
        spectral_parameters = {
            key: (np.expand_dims(value, axis=0) if value.ndim == 1 else value)
            for key, value in self.spectral_parameters.items()
        }

        #Assuming delta frequencies
        if bandpass is None:
            if freq.size == 1:
                scaling = self.get_freq_scaling(
                    freq, freq_ref, **spectral_parameters
                )
            else:
                scaling = (
                    self.get_freq_scaling(freq, freq_ref, **spectral_parameters)
                    for freq in freq
                )
            if self.diffuse:
                emission = amp*scaling
                if fwhm.value != 0.0:
                    if self.is_polarized:
                        emission = hp.smoothing(
                            emission, fwhm=fwhm.to(u.rad).value
                        )*emission.unit
                    else:
                        emission[0] = hp.smoothing(
                            emission[0], fwhm=fwhm.to(u.rad).value
                        )*emission.unit    

            else:
                # self.amp is not a healpix map for non diffuse comps
                emission = self.get_map(amp=amp*scaling, fwhm=fwhm)
            
            if output_unit is not None:
                try:
                    output_unit = u.Unit(output_unit)
                    emission = emission.to(
                        output_unit, equivalencies=u.brightness_temperature(freq)
                    )
                except ValueError:
                    if output_unit.lower().endswith('k_rj'):
                        output_unit = u.Unit(output_unit[:-3])
                    elif output_unit.lower().endswith('k_cmb'):
                        output_unit = u.Unit(output_unit[:-4])   
                        emission *= brightness_to_thermodynamical(freq)
                
            return emission

        # Perform bandpass integration
        else:
            bandpass = get_normalized_bandpass(bandpass, freq, input_unit)
            bandpass_coefficient = get_bandpass_coefficient(
                bandpass, freq, output_unit
            )
            bandpass_scaling = self._get_bandpass_scaling(
                freq, bandpass, spectral_parameters
            )

            if self.diffuse:
                emission = amp*bandpass_scaling*bandpass_coefficient
                if fwhm.value != 0.0:
                    if self.is_polarized:
                        emission = hp.smoothing(
                            emission, fwhm=fwhm.to(u.rad).value
                        )*emission.unit
                    else:
                        emission[0] = hp.smoothing(
                            emission[0], fwhm=fwhm.to(u.rad).value
                        )*emission.unit   

            else:
                emission = self.get_map(
                    amp=amp*bandpass_scaling, fwhm=fwhm
                )*bandpass_coefficient

            return emission.to(_get_astropy_unit(output_unit))


    def _get_bandpass_scaling(self, freqs, bandpass, spectral_parameters):
        """Returns the frequency scaling factor given a bandpass profile and a
        corresponding frequency array.

        Args:
        -----
        freqs (`astropy.units.Quantity`):
            Bandpass profile frequencies.
        bandpass (`astropy.units.Quantity`):
            Normalized bandpass profile.

        Returns:
        --------
        float, `numpy.ndarray`
            Frequency scaling factor given a bandpass.

        """
        interp_parameters = get_interp_parameters(spectral_parameters)

        # Component does not have any spatially varying spectral parameters
        if not interp_parameters:
            freq_scaling = self.get_freq_scaling(
                freqs, self.freq_ref, **spectral_parameters
            )
            # Reshape to support broadcasting for comps where freq_ref = None 
            # e.g cmb
            if freq_scaling.ndim > 1:
                return np.expand_dims(
                    np.trapz(freq_scaling*bandpass, freqs), axis=1
                )
            return np.trapz(freq_scaling*bandpass, freqs)

        # Component has one sptatially varying spectral parameter
        elif len(interp_parameters) == 1:
            return interp1d(
                self, freqs, bandpass, interp_parameters, 
                spectral_parameters.copy()
            )

        # Component has two sptatially varying spectral parameter
        elif len(interp_parameters) == 2:    
            return interp2d(
                self, freqs, bandpass, interp_parameters, 
                spectral_parameters.copy()
            )

        else:
            raise NotImplementedError(
                'Bandpass integration for comps with more than two spectral '
                'parameters is not implemented'
            )


    def to_nside(self, new_nside):
        """ud_grades all maps in the component to a new nside.

        Args:
        -----
        new_nside (int):
            Healpix map resolution parameter.

        """
        if not self.diffuse:
            return

        nside = hp.get_nside(self.amp)
        if new_nside == nside:
            print(f'Model is already at nside {nside}')
            return
        if not hp.isnsideok(new_nside, nest=True):
            raise ValueError(f'nside: {new_nside} is not valid.')

        self.amp = hp.ud_grade(self.amp.value, new_nside)*self.amp.unit
        for key, val in self.spectral_parameters.items():
            if hp.nside2npix(nside) in np.shape(val):
                try:
                    self.spectral_parameters[key] = hp.ud_grade(val.value, 
                                                            new_nside)*val.unit
                except AttributeError:
                    self.spectral_parameters[key] = hp.ud_grade(val, new_nside)


    @property
    def is_polarized(self):
        """Returns True if component is polarized and False if not"""
        if self.amp.shape[0] == 3:
            return True
        return False


    def __repr__(self):
        main_repr = f'{self.__class__.__name__}'
        main_repr += '('
        extra_repr = ''
        for key in self.spectral_parameters.keys():
            extra_repr += f'{key}, '
        if extra_repr:
            extra_repr = extra_repr[:-2]
        main_repr += extra_repr
        main_repr += ')'

        return main_repr



class Synchrotron(Component):
    r"""Synchrotron component class. Defined using the convention in 
    `BeyondPlanck (2020), Section 3.3.1 <https://arxiv.org/pdf/2011.05609.pdf>`_;

    .. math::

        \boldsymbol{s}_\mathrm{RJ}^{\mathrm{s}}(\nu) \propto
        \left( \frac{\nu}{\nu_\mathrm{0,s}} \right)^
        {\beta + C \ln \nu / \nu_{0,s}}.

    This is a generic power law given at a reference frequency :math:`\nu_{s,0}`
    with a power law :math:`\beta` in Rayleigh-Jeans temperature.
    :math:`C` is set to 0 for all current implementations as of BP9.

    Parameters
    ----------
    amp : `astropy.units.Quantity`
        Emission templates of the component at the reference frequencies given
        by freq_ref.
    freq_ref : `astropy.units.Quantity`
        Reference frequencies :math:`\nu_\mathrm{0,s}` for the amplitude 
        template in units of GHz. Shape is either (1,) or (3, 1)
    beta : `numpy.ndarray`, `astropy.units.Quantity`
        The power law spectral index :math:`\beta`. The spectral index can 
        vary over the sky, and is therefore commonly given as a 
        shape (3, `npix`) array, but it can take the value of a scalar.

    Attributes
    ----------
    label : str
        Component label.
    diffuse : bool
        Whether or not the component is diffuse in nature.
    amp : `astropy.units.Quantity`
        Emission templates of synchrotron at the reference frequencies given
        by `freq_ref`.
    freq_ref : `astropy.units.Quantity`
        Reference frequencies :math:`\nu_\mathrm{0,s}` for the amplitude 
        template in units of GHz.
    spectral_parameters : dict
        Dictionary containing the spectral parameters.

    Methods
    -------
    get_freq_scaling    
    __call__
    """

    label = 'synch'
    diffuse = True


    def __init__(self, amp, freq_ref, beta):
        super().__init__(amp, freq_ref, beta=beta)


    def get_freq_scaling(self, freq, freq_ref, beta):
        r"""Computes the frequency scaling :math:`f_{\mathrm{s}}(\nu)` 
        from the reference frequency :math:`\nu_\mathrm{0,s}` to a frequency 
        :math:`\nu`, 

        .. math::

            f_{\mathrm{s}}(\nu) = \left( \frac{\nu}{\nu_\mathrm{0,s}} \right)
            ^{\beta}.

        Parameters
        ----------
        freq : `astropy.units.Quantity`)
            Frequency at which to evaluate the model.
        freq_ref : `astropy.units.Quantity`
            Reference frequencies for the amplitude map.
        beta : `numpy.ndarray`, `astropy.units.Quantity`
            The power law spectral index.
            
        Returns
        -------
        scaling : `astropy.units.Quantity`
            Frequency scaling factor with dimensionless units.

        """
        scaling = (freq/freq_ref)**beta
        return scaling



class Dust(Component):
    r"""Thermal dust component class. Defined using the convention in 
    `BeyondPlanck (2020), Section 3.3.3 <https://arxiv.org/pdf/2011.05609.pdf>`_;

    .. math::

        \boldsymbol{s}_\mathrm{RJ}^{\mathrm{d}}(\nu) \propto 
        \frac{\nu^{\beta_{\mathrm{d}}+1}}{\mathrm{e}^{h\nu/kT_{\mathrm{d}}}-1}.

    This is a modified blackbody with a power law spectral index :math:`\beta` 
    in Rayleigh-Jeans temperature, and thermal dust temperature 
    :math:`T_{\mathrm{d}}`.
    

    Parameters
    ----------
    amp : `astropy.units.Quantity`
        Emission templates of the component at the reference frequencies given
        by `freq_ref`.
    freq_ref : `astropy.units.Quantity`
        Reference frequencies for the amplitude 
        template in units of GHz. Shape is either (1,) or (3, 1)
    beta : `numpy.ndarray`, `astropy.units.Quantity`
        The power law spectral index :math:`\beta`. The spectral index can vary 
        over the sky, and is therefore commonly given as a shape (3, `npix`) 
        array, but it can take the value of a scalar.
    T : `astropy.units.Quantity`:
        Temperature of the blackbody with unit :math:`\mathrm{K}_\mathrm{RJ}`.
        Can be a single value or a map with shape (`npix`,).

    Attributes
    ----------
    label : str
        Component label.
    diffuse : bool
        Whether or not the component is diffuse in nature.
    amp : `astropy.units.Quantity`
        Emission templates of synchrotron at the reference frequencies given
        by `freq_ref`.
    freq_ref : `astropy.units.Quantity`
        Reference frequencies :math:`\nu_\mathrm{0,d}` for the amplitude 
        template in units of GHz.
    spectral_parameters : dict
        Dictionary containing the spectral parameters.

    Methods
    -------
    get_freq_scaling    
    __call__
    """

    label = 'dust'
    diffuse = True


    def __init__(self, amp, freq_ref, beta, T):
        super().__init__(amp, freq_ref, beta=beta, T=T)


    def get_freq_scaling(self, freq, freq_ref, beta, T):
        r"""Computes the frequency scaling :math:`f_{\mathrm{d}}(\nu)` from the 
        reference frequency :math:`\nu_\mathrm{0,d}` to a frequency 
        :math:`\nu`, given the spectral index :math:`\beta` and the 
        electron temperature :math:`T_\mathrm{d}`.

        .. math::

            f_{\mathrm{d}}(\nu) = \left( \frac{\nu}{\nu_\mathrm{0,d}} \right)
            ^{\beta-2}\frac{B_\nu(T_{\mathrm{d}})}
            {B_{\nu_\mathrm{0,d}}(T_{\mathrm{d}})},

        where :math:`B_\nu(T_\mathrm{d})` is the blackbody emission.

        Parameters
        ----------
        freq : `astropy.units.Quantity`)
            Frequency at which to evaluate the model.
        freq_ref : `astropy.units.Quantity`
            Reference frequencies for the amplitude map.
        beta : `numpy.ndarray`, `astropy.units.Quantity`
            The power law spectral index.
        T : `astropy.units.Quantity`
            Temperature of the blackbody.
            
        Returns
        -------
        scaling : `astropy.units.Quantity`
            Frequency scaling factor with dimensionless units.
        """

        blackbody_ratio = (
            blackbody_emission(freq, T) / blackbody_emission(freq_ref, T)
        )
        scaling = (freq/freq_ref)**(beta+1) * blackbody_ratio
        return scaling



class FreeFree(Component):
    r"""Free-free component class. Defined using the convention in 
    `BeyondPlanck (2020), Section 3.3.2 <https://arxiv.org/pdf/2011.05609.pdf>`_;

    .. math::

        \boldsymbol{s}_\mathrm{RJ}^{\mathrm{ff}}(\nu) \propto
        \frac{g_{\mathrm{ff}}(T_\mathrm{e})}{\nu^2},


    where :math:`g_\mathrm{ff}` is the Gaunt factor, and :math:`T_\mathrm{e}` 
    is the electron temperature.

    Parameters
    ----------
    amp : `astropy.units.Quantity`
        Emission templates of the component at the reference frequencies given
        by freq_ref.
    freq_ref : `astropy.units.Quantity`
        Reference frequencies :math:`\nu_\mathrm{0,ff}` for the amplitude 
        template in units of GHz. Shape is either (1,) or (3, 1)
    Te : `astropy.units.Quantity`
        Electron temperature map with unit K.

    Attributes
    ----------
    label : str
        Component label.
    diffuse : bool
        Whether or not the component is diffuse in nature.
    amp : `astropy.units.Quantity`
        Emission templates of synchrotron at the reference frequencies given
        by `freq_ref`.
    freq_ref : `astropy.units.Quantity`
        Reference frequencies :math:`\nu_\mathrm{0,\mathrm{ff}}` for the amplitude 
        template in units of GHz.
    spectral_parameters : dict
        Dictionary containing the spectral parameters.

    Methods
    -------
    get_freq_scaling    
    __call__
    """

    label = 'ff'
    diffuse = True


    def __init__(self, amp, freq_ref, Te):
        super().__init__(amp, freq_ref, Te=Te)


    def get_freq_scaling(self, freq, freq_ref, Te):
        r"""Computes the frequency scaling :math:`f_{\mathrm{ff}}(\nu)` from the 
        reference frequency :math:`\nu_{0, \mathrm{ff}}` to a frequency 
        :math:`\nu`, given the electron temperature :math:`T_\mathrm{e}`.

        .. math::

            f_{\mathrm{ff}}(\nu) = \frac{g_{\mathrm{ff}}\left(\nu ; T_{e}\right)}
            {g_{\mathrm{ff}}\left(\nu_{0, \mathrm{ff}} ; T_{e}\right)}
            \left(\frac{\nu_{0, \mathrm{ff}}}{\nu}\right)^{2}.

        Parameters
        ----------
        freq : `astropy.units.Quantity`
            Frequency at which to evaluate the model.
        freq_ref : `astropy.units.Quantity`
            Reference frequencies for the amplitude map.
        Te : `astropy.units.Quantity`
            Electron temperature.
            
        Returns
        -------
        scaling : `astropy.units.Quantity`
            Frequency scaling factor with dimensionless units.
        """

        gaunt_factor_ratio = gaunt_factor(freq, Te) / gaunt_factor(freq_ref, Te)
        scaling = (freq_ref/freq)**2 * gaunt_factor_ratio
        return scaling



class AME(Component):
    r"""Spinning dust component class. Defined using the convention in 
    `BeyondPlanck (2020), Section 3.3.4 <https://arxiv.org/pdf/2011.05609.pdf>`_;

    .. math::


        \mathbf{s}_{\mathrm{RJ}}^{\mathrm{sd}}(\nu) \propto 
        \nu^{-2} s_{0}^{\mathrm{sd}}\left(\nu \cdot 
        \frac{30.0 \mathrm{GHz}}{\nu_{p}}\right)


    where the peak frequency :math:`\nu_p` is 30 GHz.

    Parameters
    ----------
    amp : `astropy.units.Quantity`
        Emission templates of the component at the reference frequencies given
        by freq_ref.
    freq_ref : `astropy.units.Quantity`
        Reference frequencies :math:`\nu_\mathrm{0,sd}` for the amplitude 
        template in units of GHz. Shape is either (1,) or (3, 1)
    nu_p : `astropy.units.Quantity`
        Peak frequency.

    Attributes
    ----------
    label : str
        Component label.
    diffuse : bool
        Whether or not the component is diffuse in nature.
    amp : `astropy.units.Quantity`
        Emission templates of synchrotron at the reference frequencies given
        by `freq_ref`.
    freq_ref : `astropy.units.Quantity`
        Reference frequencies :math:`\nu_\mathrm{0,sd}` for the amplitude 
        template in units of GHz.
    spectral_parameters : dict
        Dictionary containing the spectral parameters.

    Methods
    -------
    get_freq_scaling    
    __call__
    """

    label = 'ame'
    diffuse = True

    def __init__(self, amp, freq_ref, nu_p):
        super().__init__(amp, freq_ref, nu_p=nu_p)

        # Read in spdust2 template
        SPDUST2_FILE = DATA_DIR / 'spdust2_cnm.dat'
        spdust2_freq, spdust2_amp = np.loadtxt(SPDUST2_FILE, unpack=True)
        spdust2_freq = u.Quantity(spdust2_freq, unit=u.GHz)
        spdust2_amp = u.Quantity(spdust2_amp, unit=(u.Jy/u.sr)).to(
            u.K, equivalencies=u.brightness_temperature(spdust2_freq)
        )        
        self.spdust2 = np.array([spdust2_freq.si.value, spdust2_amp.si.value])


    def get_freq_scaling(self, freq, freq_ref, nu_p):
        r"""Computes the frequency scaling :math:`f_{\mathrm{sd}}(\nu)` from the 
        reference frequency :math:`\nu_{0, \mathrm{sd}}` to a frequency 
        :math:`\nu`, given the peak frequency :math:`\nu_p`.

        .. math::

            f_{\mathrm{sd}}(\nu) = \left(\frac{\nu_{0, \mathrm{sd}}}{\nu}\right)^{2} 
            \frac{s_{0}^{\mathrm{sd}}\left(\nu \cdot 
            \frac{\nu_{p}}{30.0 \mathrm{GHz}}\right)}
            {s_{0}^{\mathrm{sd}}\left(\nu_{0, \mathrm{sd}} \cdot \frac{\nu_{p}}
            {30.0 \mathrm{GHz}}\right)}  


        Parameters
        ----------
        freq : `astropy.units.Quantity`
            Frequency at which to evaluate the model.
        freq_ref : `astropy.units.Quantity`
            Reference frequencies for the amplitude map.
        nu_p : `astropy.units.Quantity`
            Peak frequency.
            
        Returns
        -------
        scaling : `astropy.units.Quantity`
            Frequency scaling factor with dimensionless units.
        """

        spdust2 = self.spdust2
        peak_scale = 30*u.GHz / nu_p

        # AME is undefined at outside of this frequency range
        if not (np.min(spdust2[0]) < ((freq*peak_scale).si.value).any() < np.max(spdust2[0])):
            return u.Quantity(0, unit=u.dimensionless_unscaled)

        interp = np.interp((freq*peak_scale).si.value, spdust2[0], spdust2[1])
        interp_ref = (
            np.interp((freq_ref*peak_scale).si.value, spdust2[0], spdust2[1])
        )
        scaling = interp/interp_ref
        return scaling



class CMB(Component):
    r"""CMB component class. Defined using the convention in 
    `BeyondPlanck (2020), Section 3.2 <https://arxiv.org/pdf/2011.05609.pdf>`_;

    .. math::

        \mathbf{s}_{\mathrm{RJ}}^{\mathrm{CMB}}(\nu) \propto \frac{x^{2} 
        \mathrm{e}^{x}}{\left(\mathrm{e}^{x}-1\right)^{2}} 
        \boldsymbol{s}^{\mathrm{CMB}}


    where :math:`x=h v / k T_{0}` and :math:`T_0 = 2.7255 \mathrm{K}` as of BP9.

    Parameters
    ----------
    amp : `astropy.units.Quantity`
        Emission templates of the component at the reference frequencies given
        by freq_ref.

    Attributes
    ----------
    label : str
        Component label.
    diffuse : bool
        Whether or not the component is diffuse in nature.
    amp : `astropy.units.Quantity`
        Emission templates of CMB in units of :math:`\mathrm{K}_{\mathrm{CMB}}´
    freq_ref : `astropy.units.Quantity`
        Reference frequency for CMB is set to ``None``.
    spectral_parameters : dict
        Dictionary containing the spectral parameters.

    Methods
    -------
    remove_dipole
    get_freq_scaling    
    __call__
    """

    label = 'cmb'
    diffuse = True

    def __init__(self, amp):
        super().__init__(amp, freq_ref=None)


    def remove_dipole(self, return_dipole=False, gal_cut=10):
        """Removes the solar dipole from the reference amplitude map.

        Parameters
        ----------
        return_dipole : bool
            If ``True``, a map of the dipole is returned. Defaut: ``False``.
        gal_cut : float
            Galactic latitude coordinate. Default: 10 degrees.

        Returns
        -------
        dipole : `astropy.units.Quantity`
            If `return_dipole` is ``True``, return the dipole map.
        """

        if not return_dipole:
            hp.remove_dipole(self.amp[0], gal_cut=gal_cut, copy=False)
        else: 
            amp_without_dipole = u.Quantity(
                hp.remove_dipole(
                    self.amp[0], gal_cut=gal_cut
                ), unit=self.amp.unit
            )
            dipole = self.amp[0] - amp_without_dipole
            self.amp[0] = amp_without_dipole

            return dipole


    def get_freq_scaling(self, freq, freq_ref=None):
        r"""Computes the frequency scaling factor :math:`f_{\mathrm{CMB}}(\nu)`. 
        For the CMB component, the frequency scaling factor is given by the 
        unit conversion factor from :math:`\mathrm{K}_\mathrm{CMB}` to 
        :math:`\mathrm{K}_\mathrm{RJ}` since the amplitude template is in 
        units of :math:`\mathrm{K}_\mathrm{CMB}`

        .. math::

            f_{\mathrm{CMB}}(\nu) = \frac{x^{2} e^{x}}{\left(e^{x}-1\right)^{2}}
    

        Parameters
        ----------
        freq : `astropy.units.Quantity`
            Frequency at which to evaluate the model.
        freq_ref : `astropy.units.Quantity`
            Reference frequencies for the amplitude map. Default: ``None``
            
        Returns
        -------
        scaling : `astropy.units.Quantity`
            Frequency scaling factor with dimensionless units.
        """

        return thermodynamical_to_brightness(freq)



class Radio(Component):
    r"""Point source component class. Defined using the convention in 
    `BeyondPlanck (2020), Section 3.4.1 <https://arxiv.org/pdf/2011.05609.pdf>`_;

    .. math::

        \mathbf{s}_{\mathrm{RJ}}^{\mathrm{src}}(\nu) \propto
        \left(\frac{\nu}{\nu_{\mathrm{0, src}}}\right)^{\alpha-2}


    Parameters
    ----------
    amp : `astropy.units.Quantity`
        Sampled amplitudes for each point source.
    freq_ref : `astropy.units.Quantity`
        Reference frequencies :math:`\nu_\mathrm{0,src}` for the point source 
        amplitudes.
    specind : `astropy.units.Quantity`
        Power law spectral index :math:`\alpha`.

    Attributes
    ----------
    label : str
        Component label.
    diffuse : bool
        Whether or not the component is diffuse in nature.
    amp : `astropy.units.Quantity`
        Point source amplitudes at the reference frequencies given
        by `freq_ref`. Note that this is not a healpix map.
    freq_ref : `astropy.units.Quantity`
        Reference frequencies :math:`\nu_\mathrm{0,\mathrm{src}}` for the amplitude 
        template in units of GHz.
    spectral_parameters : dict
        Dictionary containing the spectral parameters.

    Methods
    -------
    get_freq_scaling
    get_map
    __call__
    """

    label = 'radio'
    diffuse = False


    def __init__(self, amp, freq_ref, specind):
        super().__init__(amp, freq_ref, specind=specind)
        self.amp = u.Quantity(self.amp.value, unit='mJy')
        self.angular_coords = self._read_angular_coords()
        self.spectral_parameters['specind'] = np.squeeze(
            self.spectral_parameters['specind'][0]
        )


    def _set_nside(self, nside):
        """Imports the nside of the sky model."""
        self.nside = nside


    def _read_angular_coords(self, catalog=DATA_DIR/'radio_catalog.dat'):
        """Reads in the angular coordinates of the point sources from a given 
        catalog.

        TODO: Make sure that the correct catalog is selected for a given chain
        (in case catalogs change from run to run)

        Parameters:
        -----------
        catalog: str
            Path to the point source catalog. Default is the COM_GB6 catalog.
        
        Returns:
        --------
        coords : `numpy.ndarray`
            Longitude and latitude values of each point source. Shape is 
            (2,n_pointsources).

        """
        try:
            coords = np.loadtxt(catalog, usecols=(0,1))
        except OSError:
            raise OSError('Could not find point source catalog')

        if len(coords) == len(self.amp[0]):
            return coords
        else:
            raise ValueError('Cataloge does not match chain catalog')
        

    @u.quantity_input(amp=u.Jy, fwhm=(u.rad, u.arcmin, u.deg))
    def get_map(self, amp, nside='model_nside', fwhm=0.0*u.rad, 
                sigma=None, n_fwhm=2):
        """Maps the cataloged radio source points onto a healpix map with a 
        truncated gaussian beam. For more information, see 
        `Mitra et al. <https://arxiv.org/pdf/1005.1929.pdf>`_.

        Parameters
        ----------
        amp : `astropy.units.Quantity`
            Amplitude of the radio sources.
        nside : int
            The nside of the output map. If component is part of a sky model, 
            we automatically select the model nside. Must be >= 32 to not throw
            exception. Default: 'model_nside'.
        fwhm : `astropy.units.Quantity`
            The full width half max parameter of the Gaussian. Default: 0.0
        sigma : float
            The sigma of the Gaussian (beam radius). Overrides fwhm. 
            Default: None
        n_fwhm : int, float
            The fwhm multiplier used in computing radial cut off r_max 
            calculated as r_max = n_fwhm * fwhm of the Gaussian.
            Default: 2.
        """

        if nside == 'model_nside':
            try:
                nside = self.nside
            # Can occur when the radio component is used outside of a sky model
            except AttributeError:
                raise AttributeError(
                    'Component is not part of a sky model. Please provide an explicit nside'
                )
        
        amp = np.squeeze(amp)
        radio_template = u.Quantity(
            np.zeros(hp.nside2npix(nside)), unit=amp.unit
        )
        pix_lon, pix_lat = hp.pix2ang(
            nside, np.arange(hp.nside2npix(nside)), lonlat=True
        )
        # Point source coordinates
        angular_coords = self.angular_coords

        fwhm = fwhm.to(u.rad)
        if sigma is None:
            sigma = fwhm / (2*np.sqrt(2 * np.log(2)))

        # No smoothing nesecarry. Directly map sources to pixels
        if sigma == 0.0:
            warnings.warn('mapping point sources to pixels without beam smoothing.')
            pixels = hp.ang2pix(nside, *angular_coords.T, lonlat=True)
            beam_area = hp.nside2pixarea(nside)*u.sr
            radio_template[pixels] = amp

        # Apply gaussian (or other beam) to point source and neighboring pixels
        else:
            pix_res = hp.nside2resol(nside)
            if fwhm.value < pix_res:
                raise ValueError(
                    'fwhm must be >= pixel resolution to resolve the '
                    'point sources.'
                )
            beam_area = 2 * np.pi * sigma**2
            r_max = n_fwhm * fwhm.value

            with tqdm(total=len(angular_coords), file=sys.stdout) as pbar:
                sigma = sigma.value
                print('Smoothing point sources')

                for idx, (lon, lat) in enumerate(angular_coords):
                    vec = hp.ang2vec(lon, lat, lonlat=True)
                    inds = hp.query_disc(nside, vec, r_max)
                    r = hp.rotator.angdist(
                        np.array(
                            [pix_lon[inds], pix_lat[inds]]), np.array([lon, lat]
                        ), lonlat=True
                    )
                    radio_template[inds] += amp[idx]*gaussian_beam_2D(r, sigma)
                    pbar.update()

        radio_template = radio_template.to(
            u.uK, u.brightness_temperature(self.freq_ref, beam_area)
        )

        return np.expand_dims(radio_template, axis=0)


    def get_freq_scaling(self, freq, freq_ref, specind):
        r"""Computes the frequency scaling :math:`f_{\mathrm{src}}(\nu)` 
        from the reference frequency :math:`\nu_{0, \mathrm{src}}` to a 
        frequency :math:`\nu`, given the spectral index :math:`\alpha`.

        .. math::

            f_{\mathrm{src}}(\nu) = \sum_{j=1}^{N_{\mathrm{src}}} 
            \left(\frac{\nu}{\nu_{0, \mathrm{src}}}\right)
            ^{\alpha_{j, \mathrm{src}}-2}

        where we sum over each point source :math:`j`.

        Parameters
        ----------
        freq : `astropy.units.Quantity`
            Frequency at which to evaluate the model.
        freq_ref : `astropy.units.Quantity`
            Reference frequencies for the point source amplitudes.
        specind : `astropy.units.Quantity`
            Power law spectral index :math:`\alpha`.
            
        Returns
        -------
        scaling : `astropy.units.Quantity`
            Frequency scaling factor with dimensionless units.

        """

        scaling = (freq/freq_ref)**(specind-2)
        return scaling
